@@ -15,6 +15,21 @@
 #ifndef X86_USABILITY_H
 #define X86_USABILITY_H
 
+#include <math.h>
+#if __SSE2__
+#include <emmintrin.h>
+#if __AVX__
+#include <immintrin.h>
+#if __XOP__
+#ifdef _MSC_VER
+#include <ammintrin.h>
+#else
+#include <x86intrin.h>
+#endif
+#endif
+#endif
+#endif // __SSE2__
+
 static NCNN_FORCEINLINE signed char float2int8(float v)
 {
     int int32 = (int)round(v);
@@ -24,8 +39,6 @@ static NCNN_FORCEINLINE signed char float2int8(float v)
 }
 
 #if __SSE2__
-#include <emmintrin.h>
-
 static NCNN_FORCEINLINE float _mm_reduce_add_ps(__m128 x128)
 {
     const __m128 x64 = _mm_add_ps(x128, _mm_movehl_ps(x128, x128));
@@ -33,6 +46,47 @@ static NCNN_FORCEINLINE float _mm_reduce_add_ps(__m128 x128)
     return _mm_cvtss_f32(x32);
 }
 
+static NCNN_FORCEINLINE float _mm_reduce_max_ps(__m128 x128)
+{
+    const __m128 x64 = _mm_max_ps(x128, _mm_movehl_ps(x128, x128));
+    const __m128 x32 = _mm_max_ss(x64, _mm_shuffle_ps(x64, x64, 0x55));
+    return _mm_cvtss_f32(x32);
+}
+
+static NCNN_FORCEINLINE int _mm_reduce_add_epi32(__m128i x)
+{
+    __m128i hi64 = _mm_unpackhi_epi64(x, x);
+    __m128i sum64 = _mm_add_epi32(hi64, x);
+    __m128i hi32 = _mm_shuffle_epi32(sum64, _MM_SHUFFLE(2, 3, 0, 1));
+    __m128i sum32 = _mm_add_epi32(sum64, hi32);
+    return _mm_cvtsi128_si32(sum32);
+}
+
+static NCNN_FORCEINLINE int32_t float2int8_sse(const __m128& _v0)
+{
+    // _MM_ROUND_NEAREST round to even
+    // simulate round to nearest via +/-0.5 with round to zero
+    __m128 _p5 = _mm_set1_ps(0.5f);
+    __m128 _signmask = _mm_castsi128_ps(_mm_set1_epi32(1 << 31));
+    __m128 _sign0 = _mm_and_ps(_v0, _signmask);
+    __m128 _v0_p5 = _mm_or_ps(_p5, _sign0);
+    __m128 _v0_adj = _mm_add_ps(_v0, _v0_p5);
+    __m128i _v0_i = _mm_cvttps_epi32(_v0_adj);
+
+    __m128i _v0_s16 = _mm_packs_epi32(_v0_i, _v0_i);
+
+    _v0_s16 = _mm_min_epi16(_v0_s16, _mm_set1_epi16(127));
+    _v0_s16 = _mm_max_epi16(_v0_s16, _mm_set1_epi16(-127));
+
+    __m128i _v8 = _mm_packs_epi16(_v0_s16, _v0_s16);
+
+#if defined(__x86_64__) || defined(_M_X64)
+    return (int32_t)_mm_cvtsi128_si64(_v8);
+#else
+    return _mm_cvtsi128_si32(_v8);
+#endif
+}
+
 static NCNN_FORCEINLINE int64_t float2int8_sse(const __m128& _v0, const __m128& _v1)
 {
     // _MM_ROUND_NEAREST round to even
@@ -55,10 +109,13 @@ static NCNN_FORCEINLINE int64_t float2int8_sse(const __m128& _v0, const __m128&
 
     __m128i _v8 = _mm_packs_epi16(_v01_s16, _v01_s16);
 
-    // TODO use _mm_cvtsi128_si64 on 64bit target
+#if defined(__x86_64__) || defined(_M_X64)
+    return _mm_cvtsi128_si64(_v8);
+#else
     int64_t v8[2];
     _mm_storeu_si128((__m128i*)v8, _v8);
     return v8[0];
+#endif
 }
 
 static NCNN_FORCEINLINE __m128i float2int8_sse(const __m128& _v0, const __m128& _v1, const __m128& _v2, const __m128& _v3)
@@ -96,45 +153,57 @@ static NCNN_FORCEINLINE __m128i float2int8_sse(const __m128& _v0, const __m128&
 
     return _v8;
 }
-#if __SSE2__
-#ifndef __AVX2__
 
+#ifndef __FMA__
 static NCNN_FORCEINLINE __m128 _mm_comp_fmadd_ps(__m128 _a, const __m128 _b, const __m128 _c)
 {
     return _mm_add_ps(_mm_mul_ps(_a, _b), _c);
 }
-#endif
-#endif
-#if __AVX__
-#include <immintrin.h>
-#ifndef __AVX2__
-static NCNN_FORCEINLINE __m256 _mm256_comp_fmadd_ps(__m256 _a, const __m256 _b, const __m256 _c)
+static NCNN_FORCEINLINE __m128 _mm_comp_fnmadd_ps(__m128 _a, const __m128 _b, const __m128 _c)
 {
-    return _mm256_add_ps(_mm256_mul_ps(_a, _b), _c);
+    return _mm_sub_ps(_c, _mm_mul_ps(_a, _b));
 }
-#ifndef __SSE2__
-static NCNN_FORCEINLINE __m128 _mm_comp_fmadd_ps(__m128 _a, const __m128 _b, const __m128 _c)
-{
-    return _mm_add_ps(_mm_mul_ps(_a, _b), _c);
-}
-#endif
 #else
 static NCNN_FORCEINLINE __m128 _mm_comp_fmadd_ps(__m128 _a, const __m128 _b, const __m128 _c)
 {
     return _mm_fmadd_ps(_a, _b, _c);
 }
+static NCNN_FORCEINLINE __m128 _mm_comp_fnmadd_ps(__m128 _a, const __m128 _b, const __m128 _c)
+{
+    // return -a * b + c
+    return _mm_fnmadd_ps(_a, _b, _c);
+}
+#endif // !__FMA__
+
+#if __AVX__
+#ifndef __FMA__
+static NCNN_FORCEINLINE __m256 _mm256_comp_fmadd_ps(__m256 _a, const __m256 _b, const __m256 _c)
+{
+    return _mm256_add_ps(_mm256_mul_ps(_a, _b), _c);
+}
+static NCNN_FORCEINLINE __m256 _mm256_comp_fnmadd_ps(__m256 _a, const __m256 _b, const __m256 _c)
+{
+    return _mm256_sub_ps(_c, _mm256_mul_ps(_a, _b));
+}
+#else
 static NCNN_FORCEINLINE __m256 _mm256_comp_fmadd_ps(__m256 _a, const __m256 _b, const __m256 _c)
 {
     return _mm256_fmadd_ps(_a, _b, _c);
 }
+static NCNN_FORCEINLINE __m256 _mm256_comp_fnmadd_ps(__m256 _a, const __m256 _b, const __m256 _c)
+{
+    // return -a * b + c
+    return _mm256_fnmadd_ps(_a, _b, _c);
+}
 #endif
-#if __AVX2__
 
+#if __AVX2__
 static NCNN_FORCEINLINE __m256 loadfp16(const unsigned short* ptr)
 {
     return _mm256_cvtph_ps(_mm_lddqu_si128((__m128i*)(ptr)));
 }
 #endif
+
 static NCNN_FORCEINLINE __m256 _mm256_fmadd_1_ps(__m256 a, __m256 b, float c)
 {
     return _mm256_comp_fmadd_ps(b, _mm256_set1_ps(c), a);
@@ -142,7 +211,8 @@ static NCNN_FORCEINLINE __m256 _mm256_fmadd_1_ps(__m256 a, __m256 b, float c)
 
 static NCNN_FORCEINLINE __m256 _mm256_fmrsub_1_ps(__m256 a, __m256 b, float c)
 {
-    return _mm256_sub_ps(a, _mm256_mul_ps(b, _mm256_set1_ps(c)));
+    // return a - b * c
+    return _mm256_comp_fnmadd_ps(b, _mm256_set1_ps(c), a);
 }
 // From: https://stackoverflow.com/a/25627536
 static NCNN_FORCEINLINE void transpose8_ps(__m256& row0, __m256& row1, __m256& row2, __m256& row3, __m256& row4, __m256& row5, __m256& row6, __m256& row7)
@@ -224,6 +294,14 @@ static NCNN_FORCEINLINE float _mm256_reduce_add_ps(__m256 x)
     return _mm_cvtss_f32(x32);
 }
 
+static NCNN_FORCEINLINE float _mm256_reduce_max_ps(__m256 x)
+{
+    const __m128 x128 = _mm_max_ps(_mm256_extractf128_ps(x, 1), _mm256_castps256_ps128(x));
+    const __m128 x64 = _mm_max_ps(x128, _mm_movehl_ps(x128, x128));
+    const __m128 x32 = _mm_max_ss(x64, _mm_shuffle_ps(x64, x64, 0x55));
+    return _mm_cvtss_f32(x32);
+}
+
 static NCNN_FORCEINLINE int64_t float2int8_avx(const __m256& _v0)
 {
     // _MM_FROUND_TO_NEAREST_INT round to even
@@ -252,10 +330,13 @@ static NCNN_FORCEINLINE int64_t float2int8_avx(const __m256& _v0)
 
     __m128i _v8 = _mm_packs_epi16(_v01_s16low, _v01_s16low);
 
-    // TODO use _mm_cvtsi128_si64 on 64bit target
+#if defined(__x86_64__) || defined(_M_X64)
+    return _mm_cvtsi128_si64(_v8);
+#else
     int64_t v8[2];
     _mm_storeu_si128((__m128i*)v8, _v8);
     return v8[0];
+#endif
 }
 
 static NCNN_FORCEINLINE __m128i float2int8_avx(const __m256& _v0, const __m256& _v1)
@@ -326,6 +407,97 @@ static NCNN_FORCEINLINE void _mm256_comp_fmadd_ps8(__m256& _sum,
     _mm256_comp_fmadd_ps4(_sum, _w4, _w5, _w6, _w7, _v4, _v5, _v6, _v7);
 }
 
+#if __AVX512F__
+static NCNN_FORCEINLINE void transpose16_ps(__m512& _r0, __m512& _r1, __m512& _r2, __m512& _r3, __m512& _r4, __m512& _r5, __m512& _r6, __m512& _r7,
+        __m512& _r8, __m512& _r9, __m512& _ra, __m512& _rb, __m512& _rc, __m512& _rd, __m512& _re, __m512& _rf)
+{
+    __m512 _tmp0 = _mm512_unpacklo_ps(_r0, _r1);
+    __m512 _tmp1 = _mm512_unpackhi_ps(_r0, _r1);
+    __m512 _tmp2 = _mm512_unpacklo_ps(_r2, _r3);
+    __m512 _tmp3 = _mm512_unpackhi_ps(_r2, _r3);
+    __m512 _tmp4 = _mm512_unpacklo_ps(_r4, _r5);
+    __m512 _tmp5 = _mm512_unpackhi_ps(_r4, _r5);
+    __m512 _tmp6 = _mm512_unpacklo_ps(_r6, _r7);
+    __m512 _tmp7 = _mm512_unpackhi_ps(_r6, _r7);
+    __m512 _tmp8 = _mm512_unpacklo_ps(_r8, _r9);
+    __m512 _tmp9 = _mm512_unpackhi_ps(_r8, _r9);
+    __m512 _tmpa = _mm512_unpacklo_ps(_ra, _rb);
+    __m512 _tmpb = _mm512_unpackhi_ps(_ra, _rb);
+    __m512 _tmpc = _mm512_unpacklo_ps(_rc, _rd);
+    __m512 _tmpd = _mm512_unpackhi_ps(_rc, _rd);
+    __m512 _tmpe = _mm512_unpacklo_ps(_re, _rf);
+    __m512 _tmpf = _mm512_unpackhi_ps(_re, _rf);
+
+    __m512 _tmpg = _mm512_shuffle_ps(_tmp0, _tmp2, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmph = _mm512_shuffle_ps(_tmp0, _tmp2, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpi = _mm512_shuffle_ps(_tmp1, _tmp3, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpj = _mm512_shuffle_ps(_tmp1, _tmp3, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpk = _mm512_shuffle_ps(_tmp4, _tmp6, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpl = _mm512_shuffle_ps(_tmp4, _tmp6, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpm = _mm512_shuffle_ps(_tmp5, _tmp7, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpn = _mm512_shuffle_ps(_tmp5, _tmp7, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpo = _mm512_shuffle_ps(_tmp8, _tmpa, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpp = _mm512_shuffle_ps(_tmp8, _tmpa, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpq = _mm512_shuffle_ps(_tmp9, _tmpb, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpr = _mm512_shuffle_ps(_tmp9, _tmpb, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmps = _mm512_shuffle_ps(_tmpc, _tmpe, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpt = _mm512_shuffle_ps(_tmpc, _tmpe, _MM_SHUFFLE(3, 2, 3, 2));
+    __m512 _tmpu = _mm512_shuffle_ps(_tmpd, _tmpf, _MM_SHUFFLE(1, 0, 1, 0));
+    __m512 _tmpv = _mm512_shuffle_ps(_tmpd, _tmpf, _MM_SHUFFLE(3, 2, 3, 2));
+
+    _tmp0 = _mm512_shuffle_f32x4(_tmpg, _tmpk, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp1 = _mm512_shuffle_f32x4(_tmpo, _tmps, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp2 = _mm512_shuffle_f32x4(_tmph, _tmpl, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp3 = _mm512_shuffle_f32x4(_tmpp, _tmpt, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp4 = _mm512_shuffle_f32x4(_tmpi, _tmpm, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp5 = _mm512_shuffle_f32x4(_tmpq, _tmpu, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp6 = _mm512_shuffle_f32x4(_tmpj, _tmpn, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp7 = _mm512_shuffle_f32x4(_tmpr, _tmpv, _MM_SHUFFLE(2, 0, 2, 0));
+    _tmp8 = _mm512_shuffle_f32x4(_tmpg, _tmpk, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmp9 = _mm512_shuffle_f32x4(_tmpo, _tmps, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpa = _mm512_shuffle_f32x4(_tmph, _tmpl, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpb = _mm512_shuffle_f32x4(_tmpp, _tmpt, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpc = _mm512_shuffle_f32x4(_tmpi, _tmpm, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpd = _mm512_shuffle_f32x4(_tmpq, _tmpu, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpe = _mm512_shuffle_f32x4(_tmpj, _tmpn, _MM_SHUFFLE(3, 1, 3, 1));
+    _tmpf = _mm512_shuffle_f32x4(_tmpr, _tmpv, _MM_SHUFFLE(3, 1, 3, 1));
+
+    _r0 = _mm512_shuffle_f32x4(_tmp0, _tmp1, _MM_SHUFFLE(2, 0, 2, 0));
+    _r1 = _mm512_shuffle_f32x4(_tmp2, _tmp3, _MM_SHUFFLE(2, 0, 2, 0));
+    _r2 = _mm512_shuffle_f32x4(_tmp4, _tmp5, _MM_SHUFFLE(2, 0, 2, 0));
+    _r3 = _mm512_shuffle_f32x4(_tmp6, _tmp7, _MM_SHUFFLE(2, 0, 2, 0));
+    _r4 = _mm512_shuffle_f32x4(_tmp8, _tmp9, _MM_SHUFFLE(2, 0, 2, 0));
+    _r5 = _mm512_shuffle_f32x4(_tmpa, _tmpb, _MM_SHUFFLE(2, 0, 2, 0));
+    _r6 = _mm512_shuffle_f32x4(_tmpc, _tmpd, _MM_SHUFFLE(2, 0, 2, 0));
+    _r7 = _mm512_shuffle_f32x4(_tmpe, _tmpf, _MM_SHUFFLE(2, 0, 2, 0));
+    _r8 = _mm512_shuffle_f32x4(_tmp0, _tmp1, _MM_SHUFFLE(3, 1, 3, 1));
+    _r9 = _mm512_shuffle_f32x4(_tmp2, _tmp3, _MM_SHUFFLE(3, 1, 3, 1));
+    _ra = _mm512_shuffle_f32x4(_tmp4, _tmp5, _MM_SHUFFLE(3, 1, 3, 1));
+    _rb = _mm512_shuffle_f32x4(_tmp6, _tmp7, _MM_SHUFFLE(3, 1, 3, 1));
+    _rc = _mm512_shuffle_f32x4(_tmp8, _tmp9, _MM_SHUFFLE(3, 1, 3, 1));
+    _rd = _mm512_shuffle_f32x4(_tmpa, _tmpb, _MM_SHUFFLE(3, 1, 3, 1));
+    _re = _mm512_shuffle_f32x4(_tmpc, _tmpd, _MM_SHUFFLE(3, 1, 3, 1));
+    _rf = _mm512_shuffle_f32x4(_tmpe, _tmpf, _MM_SHUFFLE(3, 1, 3, 1));
+}
+
+static NCNN_FORCEINLINE float _mm512_comp_reduce_add_ps(__m512 x)
+{
+    const __m256 x256 = _mm256_add_ps(_mm512_castps512_ps256(x), _mm512_extractf32x8_ps(x, 1));
+    const __m128 x128 = _mm_add_ps(_mm256_castps256_ps128(x256), _mm256_extractf128_ps(x256, 1));
+    const __m128 x64 = _mm_add_ps(x128, _mm_movehl_ps(x128, x128));
+    const __m128 x32 = _mm_add_ss(x64, _mm_shuffle_ps(x64, x64, 0x55));
+    return _mm_cvtss_f32(x32);
+}
+
+static NCNN_FORCEINLINE float _mm512_comp_reduce_max_ps(__m512 x)
+{
+    const __m256 x256 = _mm256_max_ps(_mm512_castps512_ps256(x), _mm512_extractf32x8_ps(x, 1));
+    const __m128 x128 = _mm_max_ps(_mm256_castps256_ps128(x256), _mm256_extractf128_ps(x256, 1));
+    const __m128 x64 = _mm_max_ps(x128, _mm_movehl_ps(x128, x128));
+    const __m128 x32 = _mm_max_ss(x64, _mm_shuffle_ps(x64, x64, 0x55));
+    return _mm_cvtss_f32(x32);
+}
+#endif // __AVX512F__
 #endif // __AVX__
 #endif // __SSE2__
 
